require 'levdon/version'

require 'json'          # default
require 'securerandom'  # default
require 'fileutils'     # default
require 'stringio'      # default
require 'open-uri'      # default

require 'rmagick'                 # sudo gem install rmagick
require 'net/http/post/multipart' # sudo gem install multipart-post #https://github.com/nicksieger/multipart-post

require 'tempfile'
require 'streamio-ffmpeg' # gem install streamio-ffmpeg / brew install ffmpeg
#require 'filemagic' # gem install ruby-filemagic / brew install libmagick


module Levdon
  APPLICATION_ID  = "Levdon"
  API_URL         = ENV['EATER_URL'] ? ENV['EATER_URL'] : "https://api.leviadon.com/upload"
  API_VERSION     = "0.1.1"
  ENABLE_SSL      = API_URL.index("https://") ? true : false
  VUUID           = SecureRandom.uuid
  
  ONTOLOGY_LIST = {
    :CLASS_ADULT          => { :available => true,   :key => 'adult',          :desc => '' },
    :CLASS_VIOLENCE       => { :available => false,  :key => 'violence',       :desc => 'closed beta does not support a this class.' },
    :CLASS_GENERAL_OBJECT => { :available => false,  :key => 'general_object', :desc => 'closed beta does not support a this class.' },
    :CLASS_MEDICAL        => { :available => false,  :key => 'medical',        :desc => 'closed beta does not support a this class.' },
    :CLASS_MEAL           => { :available => false,  :key => 'meal',           :desc => 'closed beta does not support a this class.' },
    :CLASS_3D             => { :available => false,  :key => '3D',             :desc => 'closed beta does not support a this class.' },
    :CLASS_2D             => { :available => false,  :key => '2D',             :desc => 'closed beta does not support a this class.' },
    :CLASS_FACE           => { :available => false,  :key => 'face',           :desc => 'closed beta does not support a this class.' },
    :CLASS_BODY           => { :available => false,  :key => 'body',           :desc => 'closed beta does not support a this class.' },
    :CLASS_POSE           => { :available => false,  :key => 'pose',           :desc => 'closed beta does not support a this class.' },
    :CLASS_ANIMAL         => { :available => false,  :key => 'animal',         :desc => 'closed beta does not support a this class.' },
    :CLASS_OCR            => { :available => false,  :key => 'ocr',            :desc => 'closed beta does not support a this class.' }
  }
  
  ONTOLOGY_REV_LIST = {}
  ONTOLOGY_LIST.each{|k,v| ONTOLOGY_REV_LIST[v[:key]] = k }

  
  class NonBlockLineStream
    def initialize(input_stream, output_stream)
      @input_stream = input_stream
      @output_stream = output_stream
      @read_queue = []
      @write_queue = []
      @rlnb_buffer = ""
    end
  
    def read
      if(@read_queue.length)
        return @read_queue.shift
      end
      return nil
    end
  
    def write(s)
      @write_queue.push(s)
      begin
        wrote_len = @output_stream.write_nonblock(@write_queue[0])
        if(@write_queue[0].length == wrote_len)
          @write_queue.shift
        else
          @write_queue[0] = @write_queue[0].slice(wrote_len,@write_queue[0].length)
        end
      rescue => e
        puts "Write error"
        puts e.class
        puts e.message
        puts e.backtrace
      end
      
      @write_queue.length
    end
  
    def poll
      has_event = false
      begin
        if(@io)
          if(@write_queue.length > 0)
            wrote_len = @output_stream.write_nonblock(@write_queue[0])
            if(@write_queue[0].length = wrote_len)
              @write_queue.shift
            else
              @write_queue[0] = @write_queue[0].slice(wrote_len,@write_queue[0].length)
            end
            has_event = true
          end
        end
      rescue => e
        puts "Write error"
        puts e.class
        puts e.message
        puts e.backtrace
      end
    
      begin
        stream = @input_stream.read_nonblock(8192)
        @rlnb_buffer << stream
        if(@rlnb_buffer.index("\n"))
          sp = @rlnb_buffer.split("\n")
          last = ""
          if(sp.length > 1)
            last = sp.pop
          end
          sp.each{|e|
            @read_queue.push(e)
          }
          @rlnb_buffer = last
        end
        # while ch = @input_stream.read_nonblock(1)
        #   @rlnb_buffer << ch
        #   if ch == "\n" then
        #     @read_queue.push(@rlnb_buffer)
        #     @rlnb_buffer = ""
        #   end
        # end
      rescue Errno::EAGAIN
      rescue => e
        puts "Read error"
        puts e.class
        puts e.message
        puts e.backtrace
      end
      return has_event
    end
  end  

  class Worker #http://qiita.com/yuroyoro/items/92c5bc864fa9c05127a9
    attr_reader :pid

    def initialize(&block)
      @child_read, @parent_write = create_pipe
      @parent_read, @child_write = create_pipe
      @block = block
      @io_stream = NonBlockLineStream.new(@parent_read,@parent_write)
    end

    def create_pipe
      IO.pipe.map{|pipe| pipe.tap{|_| _.set_encoding("ASCII-8BIT", "ASCII-8BIT") } }
    end
  
    def run
      @pid = fork do
        @parent_read.close
        @parent_write.close
        write_to_parent(:ready)
        loop do
          args = read_from_parent
          break if args == :stop 
          result = @block.call(*args)
          write_object(result, @child_write)
        end

        @child_read.close
        @child_write.close
      end

      wait_after_fork if @pid
    end
  
    def execute(*msg)
      write_to_child(msg)
      Thread.new { read_from_child }
    end
  
    def async_execute(*msg)
      nonblock_write_to_child(msg)
    end

    def stop
      return unless alive?
      write_to_child(:stop)
      Process.wait(@pid)
    end

    def alive?
      Process.kill(0, @pid)
      true
    rescue Errno::ESRCH
      false
    end
  
    def poll
      @io_stream.poll
    end

    def write_object(obj, write)
      data = Marshal.dump(obj).gsub("\n", "@NDELIMITER@") + "\n"
      write.write data
    end

    def read_object(read)
      data = read.gets
      Marshal.load(data.chomp.gsub("@NDELIMITER@", "\n"))
    end
  
    def nonblock_read_from_child()
      data = @io_stream.read
      if(data)
        return Marshal.load(data.chomp.gsub("@NDELIMITER@", "\n"))
      end
      return nil
    end
  
    def nonblock_write_to_child(obj)
      data = Marshal.dump(obj).gsub("\n", "@NDELIMITER@") + "\n"
      @io_stream.write(data)
    end

    def read_from_child
      read_object(@parent_read)
    end

    def write_to_child(obj)
      write_object(obj, @parent_write)
    end

    def read_from_parent
      read_object(@child_read)
    end

    def write_to_parent(obj)
      write_object(obj, @child_write)
    end

    def wait_after_fork
      @child_read.close
      @child_write.close

      install_exit_handler
      install_signal_handler
  
      Thread.new {
        result = read_from_child
        raise "Failed to start worker pid #{ @pid }" unless result == :ready
        result
      }
    end

    def install_exit_handler
      at_exit do
        next unless alive?
        begin
          Process.kill("KILL", @pid)
          Process.wait(@pid)
        rescue Errno::ESRCH
          # noop
        rescue => e
          puts "error at_exit: #{ e }"
          raise e
        end
      end
    end

    def install_signal_handler
      [:INT, :QUIT].each do |signal|
        old_handler = Signal.trap(signal) {
          Process.kill(signal, @pid)
          Process.wait(@pid)
          old_handler.call
        }
      end
    end
  end

  def self.identify_image(path,&block)
    proc = lambda() {|e|      
      unless File.directory?(e)
        begin
          fname = File.basename(e)
          # TODO
          if(fname.index("@STANDARDIZED@_") == 0 or fname.index("@Q@_") == 0)
            return nil
          end
          
          fsize = File.binread(e).size
          if(fsize > 1024*1024*5)
            obj = {}
            obj[:source_path] = e
            obj[:source_filename] = File.basename(e)
            obj[:error] = "Too big file :" + fsize.to_s
            return obj
          end
          if(fsize < 1024*10)
            obj = {}
            obj[:source_path] = e
            obj[:source_filename] = File.basename(e)
            obj[:error] = "Too small file :" + fsize.to_s
            return obj
          end
          
          time = Time.new
          img = Magick::Image.from_blob( File.read(e) ).shift
          #puts (Time.new - time).to_s + " : " + fsize.to_s + " : "+e
          if %w(JPEG GIF PNG).member?(img.format)
            obj = {}
            obj[:source_path] = e
            obj[:source_filename] = File.basename(e)
            obj[:image] = img;
            obj[:time] = Time.new - time
            obj[:size] = fsize
            return obj
          else
            obj = {}
            obj[:source_path] = e
            obj[:source_filename] = File.basename(e)
            obj[:image] = img;
            obj[:error] = "Unknown file"
            return obj
          end        
            
        rescue => err
          puts "Warning: cought a something error."
          obj = {}
          obj[:source_path] = e
          obj[:source_filename] = File.basename(e)
          obj[:image] = img;
          obj[:error] = err
          return obj
        end
      end
      return nil
    }
    unless File.directory?((path))
      obj = proc.call(path)
      block.call(obj) if(obj)
    else
      Dir.glob(path + "/" + "**/*").each{|e|
        obj = proc.call(e)
        block.call(obj) if(obj)
      }
    end
  end

  def self.find_image(path,options={:no_check => false},&block)
    unless File.exist?(path)
      puts "Error"
      puts " File does not exist => " + path
      return
    end
    proc = lambda() {|e|
      unless File.directory?(e)
        begin
          if(options and options[:no_check] == true)
            if %w(.JPG .JPEG .GIF .PNG).member?(File.extname(e).upcase)
              obj = {}
              obj[:source_path] = e
              obj[:source_filename] = File.basename(e)
              return obj
            end
          else
            if %w(.JPG .JPEG .GIF .PNG).member?(File.extname(e).upcase)
              img = Magick::Image.from_blob( File.read(e) ).shift
              if %w(JPEG GIF PNG).member?(img.format.to_s)
                obj = {}
                obj[:source_path] = e
                obj[:source_filename] = File.basename(e)
                obj[:image] = img;
                return obj
              end
            end
          end
        rescue => e
          puts "Warning: caught a something error."
          puts e.class
          puts e.message
          puts e.backtrace
          # nothing to do
        end
      end
      return nil
    }
    unless File.directory?(path)
      obj = proc.call(path)
      block.call(obj) if(obj)
    else
      Dir.glob(path + "/" + "**/*").each{|e|
        obj = proc.call(e)
        block.call(obj) if(obj)
      }
    end
  end

  def self.high_quality_resize(stream)
    img = Magick::Image.from_blob(stream).first
    width = img.columns
    height = img.rows
    if(width > 512 and height > 512)
      img = img.resize_to_fit(512, 512)
    end
    img.to_blob {
      self.quality = 90
      self.format = "jpeg"
    }
  end

  def self.low_quality_resize(stream)
    img = Magick::Image.from_blob(stream).first
    img = img.resize(224,224)
    img.to_blob {
      self.quality = 20
      self.format = "jpeg"
    }
  end

  def self.prob_resize(stream)
    if(Random.rand() < 0.9)
      return low_quality_resize(stream)
    end
    return high_quality_resize(stream)
  end


  def self.standardize(path,options={:mode=>:delete})
    mode = options[:mode]
    dest = options[:dest]
    unless(mode)
      puts "Error"
      puts "Require parameter"
      puts " Specify a :mode => :delete or :move or :rename"
      exit(9)
    end
    unless(%w{delete delete rename}.member?(mode.to_s))
      puts "Error"
      puts "Require parameter"
      puts " Specify a :mode => :delete or :move or :rename"
      exit(9)
    end
    
    dirname = path
    unless File.directory?(path)
      dirname = File.dirname(path)
    end
    puts "Entry path : " + path
    puts "Destination : " + dirname


    identify_image(path) {|e|
      source_path     = e[:source_path]
      source_filename = e[:source_filename];
      dst_dirname     = File.dirname(source_path)
      if(e[:error])
        File.delete(source_path) if File.exist?(source_path)
        if(mode == :delete)
          File.delete(source_path) if File.exist?(source_path)
          next
        elsif(mode == :rename)
          FileUtils.mv(source_path,File.join(dst_dirname,mark+SecureRandom.uuid+"."+source_filename.split(".")[-1]))
          next
        else
          next
        end
      end
      img             = e[:image]
      width           = img.columns
      height          = img.rows
      mark            = ""
      
      # Check already standardized file or not.
      if source_filename.index("@STANDARDIZED@") == 0 or source_filename.index("@Q@") == 0
        puts "already done : " + source_path
        next
      # else
      #   puts "processing   : " + source_path
      end
      
      # Validaiton check 
      if(width > 8192 or height > 8192)
        #puts "too big image: " + source_path
        mark = "@BIG@_"
      elsif(width < 300 or height < 300)
        #puts "thumb image  : " + source_path
        mark = "@THUMB@_"
      end
      
      puts e[:time].to_s + " : " + e[:size].to_s + " : "  + mark + " : " + source_path
      
      if(mark.length > 0)
        if(mode == :delete)
          File.delete(source_path) if File.exist?(source_path)
          next
        elsif(mode == :rename)
          FileUtils.mv(source_path,File.join(dst_dirname,mark+SecureRandom.uuid+"."+source_filename.split(".")[-1]))
          next
        elsif(mode == :move)
          if(dest)
            if(File.exist?(dest))
              if(File.directory?(dest))
                FileUtils.mv(source_path,File.join(dest,mark+SecureRandom.uuid+"."+source_filename.split(".")[-1]))
                next
              else
                puts "Error"
                puts "Destination is not directory.  => " + dest
                exit(9)
              end
            else
              puts "Error"
              puts "Destination path does not exist. => " + dest
              exit(9)
            end
          else
            puts "Error"
            puts "Destination path is nil."
            puts "You should specify a destination path for :move mode."
            exit(9)
          end
        end
      end
    
      # keeping ratio
      if(width > 1024 or height > 1024)
        img.resize_to_fit!(1024, 1024)
      end
      
      fname = source_filename.split(".")[0]
      img.format = 'JPEG'
      begin
        # Reformat and save
        img.write(File.join(dst_dirname,"@STANDARDIZED@_"+SecureRandom.uuid+".jpg")) { self.quality = 90 } # 1~100
        # Delete source file
        File.delete(source_path) if File.exist?(source_path)
      rescue => e
        puts "Write error"
        puts e.class
        puts e.message
        puts e.backtrace
      end
      # 10~30KB 80 : Q-TRAIN
      # 30~60KB 90 : Q-SOURCE
      # 50~200KB 100 : Q-SOURCE
    }
    puts "Done"
  end



  class LevdonImpl
  
    # def LevdonImpl.dealloc(workers)
    #     proc {|id|
    #       begin
    #         workers.each(&:stop)
    #         workers = nil
    #       rescue => e
    #         # nothing to do
    #       end
    #     }
    # end
    
    def initialize()
      @access_token = "DUMMY"
      @blocks       = {}
      @timeout      = 10.0 # default 10sec
      @request_cluster_num = 1
      
      # prediction parameters 
      @ontologies = [:CLASS_ADULT]
    end
  
    def _predict_impl(parameter)
      ret   = nil
      error = nil
      
      f             = parameter[:target]
      request_id    = parameter[:request_id]
      ontologies    = parameter[:ontologies]
      access_token  = parameter[:access_token]
      url           = URI.parse(API_URL)
      
      options       = {}
      options['ontologies'] = []
      error = "Class is not specified." if ontologies.length == 0
      ontologies.each{|key|
        obj = ONTOLOGY_LIST[key]
        unless obj
          error = "Invalid key => " + key.to_s 
          break
        end
        unless obj[:available]
          error = key.to_s + " class is not available.\n Reason => #{obj[:desc]}" 
          break
        end
        options['ontologies'] += [obj[:key]]
      }
      if(!error)
        stream = load(f)
        if(stream)
          io = nil
          begin
            io = StringIO.new(Levdon.prob_resize(stream))
            req = Net::HTTP::Post::Multipart.new url.path,
              "api_version"   => API_VERSION,
              "upload"        => UploadIO.new(io, "image/jpeg", "image.jpg"),
              "request_id"    => request_id,
              "access_token"  => access_token,
              "constant_id"   => APPLICATION_ID,
              "options"       => JSON.generate(options)
      
            n = Net::HTTP.new(url.host, url.port) 
            n.use_ssl = ENABLE_SSL
      
            res = n.start {|http| http.request(req) }
          
            j = JSON.parse(res.body)
            if(res.code == '200' and j['results'])
              # TODO: data struct problem
              ret = {}
              j['results'].each{|k,v|
                ret[ONTOLOGY_REV_LIST[k]] = v
              }
            else
              ret = nil
              error = j['desc']
            end
          rescue => e
            error = ""
            error += e.class.to_s + "\n"
            error += e.message.to_s + "\n"
            error += e.backtrace.to_s + "\n"
          ensure
            if(io)
              begin
                io.close()
              rescue IOError
                # nothing to do
              rescue => e
                error = ""
                error += e.class.to_s + "\n"
                error += e.message.to_s + "\n"
                error += e.backtrace.to_s + "\n"
              end
            end
          end
        else
          error = "Could not read a '#{f.to_s}'."
        end
      end
      return {:error=>error,:result=>ret}
    end
    
    def async_start(access_token,&block)
      @access_token = access_token
      @workers = @request_cluster_num.times.map{
        Worker.new{|*args|
          ret   = nil
          error = nil
          begin
            parameter   = args[0]
            parameter[:access_token] = access_token
            obj = _predict_impl(parameter)
            ret = obj[:result]
            error = obj[:error]
          rescue => e
            error = ""
            error += e.class.to_s + "\n"
            error += e.message.to_s + "\n"
            error += e.backtrace.to_s + "\n"
          end
          
          {:results => ret, :error => error}
        }
      }
      
      @workers.map(&:run).each(&:join)
      @roundrobin_counter       = 0
      @task_stacking_counter    = []
      @workers.length.times{|i| @task_stacking_counter.push([i,0]) }
      #ObjectSpace.define_finalizer(self,LevdonImpl.dealloc(@workers))
      
      # CTRL + C
      Signal.trap(:INT) {
        puts "Stop " + APPLICATION_ID + " clients "
        close()
        exit(0)
      }
      
      block.call({:error => nil})
    end
    
    def start(access_token)
      @access_token = access_token
      return {:erorr => nil}
    end
  
    def option(obj)
      obj.each{|k,v|
        if(k == :CLASS)
          @ontologies = v
        elsif(k == :PARALLEL)
          @request_cluster_num = v
        else
          puts "Invalid option : " + k.to_s
        end
      }
    end

    def detect_content_type(ouri)
      if(ouri)
        if(ouri.respond_to?(:content_type))
          ct = ouri.content_type
          if(ct)
            if(ct.index('image'))
              return 'image'
            elsif(ct.index('video'))
              return 'video'
            end
          end
        else
          return 'file'
        end
      end
      return nil
    end

    def detect_image_or_video_from_path(path)
      ext = File.extname(path).downcase
      if(ext == ".jpeg" || ext == ".png" || ext == ".gif" || ext == ".jpg" || ext == ".tiff" || ext == ".tif" || ext == ".psd" || ext == ".pdf")
        return "image"
      elsif(ext == ".mp4" || ext == ".avi" || ext == ".mov" || ext == ".3gp" || ext == ".flv" || ext == ".wmv")
        return "video"
      end
      return nil
    end
    
    def generate_thumb_from_video(f)
      thumb_url = "tmp"+Process.pid.to_s
      movie = FFMPEG::Movie.new(f)
      duration = (movie.duration / 2).floor
      path = "#{thumb_url}.jpg"
      movie.screenshot(path,  seek_time: duration)
      data = File.open(path).read()
      File.delete(path)
      return data
    end

    def to_image(fname,ct)
      if(ct == "video")
        return generate_thumb_from_video(fname)
      elsif(ct == "image")
        return File.open(fname).read
      end
      puts "Invalid media stream."
      return nil
    end
  
    def load(any)
      if(any.class.name == "String")
        if(any.index("http://") == 0 or any.index("https://") == 0)
          # from network
          ouri = open(any)
          ct = detect_content_type(ouri)
          new_data = nil
          if(ct == "video")
            new_data = generate_thumb_from_video(ouri.path)
          elsif(ct == "image")
            new_data = ouri.read
          end
          return new_data
        else
          if File::ftype(any) == "file"
            ct = detect_image_or_video_from_path(any)
            if(ct == "video")
              return generate_thumb_from_video(any)
            elsif(ct ==  "image")
              return File.open(any).read
            end
            puts "Invalid media file."
            return nil
          elsif(File::ftype(any) == "directory")
            puts "Can't load a directory. Target should be a file or URL."
            return nil
          else
            # from memory
            return any
          end
        end
      else
        # from memory
        return any
      end
      return nil
    end
  
    def poll
      @workers.each(&:poll)
      @workers.map(&:nonblock_read_from_child).each_with_index{|ret,i|
        begin
          if(ret)
            request_id = ret[:parameter][:request_id]
            key = VUUID + request_id
            if(@blocks.has_key?(key))
              @task_stacking_counter[i][1] -= 1
              @blocks[key].call(ret)
              @blocks.delete(key)
            else
              puts "Error"
              puts "Invalid response from server."
            end
          end
        rescue => e
          puts "Response error"
          puts e.class
          puts e.message
          puts e.backtrace
        end
      }
    end
  
    def enqueue(*args)
      target = @task_stacking_counter.sort{|a,b| a[1]-b[1]}[0][0]
      @task_stacking_counter[target][1] += 1
      #target = @roundrobin_counter % @workers.length
      #@roundrobin_counter += 1
      
      worker = @workers[target]
      worker.async_execute(*args)
    end
  
    def async_predict(parameter,&block)
      raise "expected key: request_id" if(parameter[:request_id])
      raise "expected key: ontologies" if(parameter[:ontologies])
      raise "required key: target. Specify a file path or URL or data." unless(parameter[:target])
      parameter[:request_id] = SecureRandom.uuid.gsub("-","")
      parameter[:ontologies] = @ontologies
      enqueue(parameter)
      @blocks[VUUID+parameter[:request_id]] = block
    end
    def queue_size
      @blocks.length
    end
  
    def close
      begin
        @workers.each(&:stop)
      rescue => e
        # nothing to do
      end
    end
  
    def predict(parameter)
      if(parameter.class.name == "String")
        parameter = {:target => parameter }
      elsif(parameter.class.name == "Hash")
        raise "expected key: request_id" if(parameter[:request_id])
        raise "expected key: ontologies" if(parameter[:ontologies])
        raise "required key: target. Specify a file path or URL or data." unless(parameter[:target])
      else
        raise "Invalid parameter type"
      end
      parameter[:request_id]    = SecureRandom.uuid.gsub("-","")
      parameter[:ontologies]    = @ontologies
      parameter[:access_token]  = @access_token
      
      ret   = nil
      error = nil
      begin
        obj         = _predict_impl(parameter)
        ret         = obj[:result]
        error       = obj[:error]
      rescue => e
        error = ""
        error += e.class.to_s + "\n"
        error += e.message.to_s + "\n"
        error += e.backtrace.to_s + "\n"
      end
      return {:results => ret, :error => error}
    end
  
  end
  
  def self.new
    LevdonImpl.new
  end
  
  def self.predict(access_token,target)
    api = LevdonImpl.new
    state = api.start(access_token)
    if(state[:error])
      raise state[:error]
    else
      result = api.predict(target)
      if(result[:error])
        raise result[:error]
      else
        return result
      end
    end
  end
end
